"""Pyvista Demos."""
from pyvista.demos.demos import (glyphs, plot_glyphs, plot_datasets,
                                 orientation_cube, orientation_plotter,
                                 plot_wave, plot_ants_plane, plot_beam)
from pyvista.demos.logo import (logo_atomized, logo_basic,
                                logo_letters, logo_voxel, plot_logo)
